/* Copyright (c) 2012 Linus Akesson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <err.h>

#include "extractgate.h"

struct sidheader {
	uint8_t			rsid;
	uint8_t			nsubtune;
	uint8_t			defsubtune;
	uint16_t		version;
	uint16_t		dataoffset;
	uint16_t		loadaddr;
	uint16_t		loadsize;
	uint16_t		initaddr;
	uint16_t		playaddr;
	char			title[32];
	char			author[32];
	char			released[32];
};

static int cycles_init = 1000000;
static int cycles_play = 20000;

int nframes = 50*60*5;

struct core ramcore;

static int readheader(struct sidheader *head, uint8_t *data, int filesize) {
	if(filesize <= 0x80) return 1;
	if(data[0] == 'P') {
		head->rsid = 0;
	} else if(data[0] == 'R') {
		head->rsid = 1;
	} else return 1;
	if(data[1] != 'S') return 1;
	if(data[2] != 'I') return 1;
	if(data[3] != 'D') return 1;
	head->version = (data[4] << 8) | data[5];
	if(head->version < 1 || head->version > 2) return 1;
	head->dataoffset = (data[6] << 8) | data[7];
	if(head->version == 1 && head->dataoffset != 0x76) return 1;
	if(head->version == 2 && head->dataoffset != 0x7c) return 1;
	head->loadaddr = (data[8] << 8) | data[9];
	if(!head->loadaddr) {
		head->loadaddr = data[head->dataoffset] | (data[head->dataoffset + 1] << 8);
		head->dataoffset += 2;
	}
	head->loadsize = filesize - head->dataoffset;
	if(head->loadaddr < 0x7e8 && head->rsid) errx(RET_RSID, "RSID standard violation");
	head->initaddr = (data[0x0a] << 8) | data[0x0b];
	if(!head->initaddr) head->initaddr = head->loadaddr;
	head->playaddr = (data[0x0c] << 8) | data[0x0d];
	head->nsubtune = (data[0x0e] << 8) | data[0x0f];
	head->defsubtune = (data[0x10] << 8) | data[0x11];
	memcpy(head->title, &data[0x16], sizeof(head->title));
	memcpy(head->author, &data[0x36], sizeof(head->author));
	memcpy(head->released, &data[0x56], sizeof(head->released));
	head->title[31] = 0;
	head->author[31] = 0;
	head->released[31] = 0;
	if(head->version > 1) {
		uint16_t flags = (data[0x76] << 8) | data[0x77];
		if(flags & 1) errx(RET_MUS, "MUS format not supported");
		if(head->rsid && (flags & 2)) errx(RET_BASIC, "BASIC tunes not supported");
		if(!head->rsid && (flags & 2)) errx(RET_PSID, "PSID tunes not supported");
	}
	return 0;
}

static void init_core(struct core *core, uint8_t *data, uint16_t loadaddr, uint16_t loadsize) {
	int i;

	memset(core->memory, 0, sizeof(core->memory));

	for(i = 0xea31; i <= 0xea86; i++) {
		core->memory[i].value = 0x68;	// pla
	}

	for(i = 0; i < loadsize; i++) {
		core->memory[loadaddr + i].value = data[i];
	}
}

static void init_tune(struct core *core, uint16_t initaddr, int tune) {
	int errcode = emulate(core, initaddr, tune, cycles_init);
	if(errcode == ERR_CYCLES) errx(1, "Max cycles exhausted during init routine. Infinite loop?");
	if(errcode) fprintf(stderr, "%s\n", emulate_err[errcode - 1]);
}

static int play_step(struct core *core, uint16_t playaddr) {
	int errcode;

	errcode = emulate(core, playaddr, 0, cycles_play);
	if(errcode == ERR_CYCLES) {
		errx(1, "Max cycles exhausted during playroutine. Infinite loop?");
	}
	if(errcode) {
		fprintf(stderr, "%s\n", emulate_err[errcode - 1]);
	}
	return errcode;
}

static int readfile(uint8_t *data, char *filename) {
	int filesize;
	FILE *f = fopen(filename, "rb");
	if(!f) err(1, "%s", filename);
	filesize = fread(data, 1, 65536, f);
	fclose(f);
	return filesize;
}

int main(int argc, char **argv) {
	uint8_t data[65536];
	int filesize;
	struct sidheader head;
	char *filename;
	int subtune;
	int i, j;
	int timespec[3] = {0, 0, 0}, timepos = 1;
	uint8_t gatestate[3];
	int rate, skip;

	if(argc != 6) {
		errx(1, "Usage: %s input.sid subtune duration rate skip", argv[0]);
	}

	filename = argv[1];
	subtune = atoi(argv[2]);
	for(i = 0; argv[3][i]; i++) {
		char ch = argv[3][i];
		if(ch >= '0' && ch <= '9') {
			timespec[timepos] *= 10;
			timespec[timepos] += ch - '0';
		} else if(ch == ':' && timepos == 1 && !timespec[0]) {
			timespec[0] = timespec[timepos];
			timespec[timepos] = 0;
		} else if(ch == '.' && timepos == 1) {
			timepos++;
		} else errx(1, "Invalid duration. Format: [min:][sec][.frames]");
	}
	nframes = ((timespec[0] * 60) + timespec[1]) * 50 + timespec[2];
	rate = atoi(argv[4]);
	skip = atoi(argv[5]);

	filesize = readfile(data, filename);
	if(readheader(&head, data, filesize)) errx(1, "Bad SID file header");
	if(subtune < 1 || subtune > head.nsubtune) errx(1, "Invalid subtune number");

	init_core(&ramcore, &data[head.dataoffset], head.loadaddr, head.loadsize);
	init_tune(&ramcore, head.initaddr, subtune - 1);

	for(i = 0; i < 3; i++) {
		gatestate[i] = 0;
	}
	for(i = 0; i < nframes; i++) {
		char buf[4] = "---";
		if(play_step(&ramcore, head.playaddr)) exit(1);
		for(j = 0; j < 3; j++) {
			if(!gatestate[j] && (ramcore.memory[0xd404 + 7 * j].value & 1)) {
				buf[j] = '+';
			}
			gatestate[j] = ramcore.memory[0xd404 + 7 * j].value & 1;
		}
		if(skip) {
			if(buf[0] != '-' || buf[1] != '-' || buf[2] != '-') {
				fprintf(stderr, "Hidden event at time %d\n", i);
			}
			skip--;
		} else {
			printf("%7d %s . . . .\n", i, buf);
			skip = rate - 1;
		}
	}

	return 0;
}
