/* Copyright (c) 2012 Linus Akesson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <err.h>
#include <string.h>
#include <math.h>

#define MAXDATA 2048

struct level {
	int	nstep;
	int	noffbeat;
	int	njump, nfreeze;
	int	peak;
	int	length;
	int	duration;
	int	pscore, qscore;
	int	pending;
	int	initdelay;
	uint8_t	data[MAXDATA];
} level[3];

int delaytable[15] = {
	96,
	95,
	48,
	47,
	36,
	35,
	24,
	23,
	18,
	17,
	12,
	11,
	6,
	5,
	1,
};

FILE *outfile;
int org;

int beatrate = 20;
int firstbeat = 0;
int maxduration = 0;

int lasttime = 0;

void putbyte(int b) {
	fputc(b, outfile);
	org++;
}

void putword(int w) {
	putbyte(w & 255);
	putbyte(w >> 8);
}

void putpadstr(char *src, int width) {
	int i;

	for(i = 0; src[i] && i < width; i++) {
		putbyte(src[i]);
	}
	for(; i < width; i++) {
		putbyte(' ');
	}
}

void putbcd(int v, int width) {
	int i, t;
	uint8_t buf[8];

	if(width > sizeof(buf)) errx(1, "BCD maxwidth");

	for(i = 0; i < width; i++) {
		t = v % 10;
		v /= 10;
		t |= (v % 10) << 4;
		v /= 10;
		buf[width - 1 - i] = t;
	}
	for(i = 0; i < width; i++) {
		putbyte(buf[i]);
	}
}

void putstepdata(struct level *l, int b) {
	if(l->length == MAXDATA) errx(1, "MAXDATA");

	l->data[l->length++] = b;
}

void putsteps(struct level *l, int bits) {
	l->pending = bits;
}

void putdelaytag(struct level *l, int i) {
	putstepdata(l, (l->pending << 4) | i);
	l->pending = 0;
}

void delay_until(struct level *l, int t) {
	int i;

	if(!l->length && !l->initdelay && t > 0 && !l->pending) {
		l->initdelay = t - lasttime;
		if(l->initdelay > 95) l->initdelay = 95;
		lasttime += l->initdelay;
		if(t == lasttime) return;
	}

	while(t - lasttime) {
		for(i = 0; i < 15; i++) {
			if(delaytable[i] <= t - lasttime) break;
		}
		if(i == 15) errx(1, "Internal error");
		putdelaytag(l, i + 1);
		lasttime += delaytable[i];
	}
	if(l->pending) {
		errx(1, "Zero delay detected");
	}
}

void readlevel(struct level *l, char *fname) {
	FILE *f;

	int now = 0;
	char buf[64], dummy[4], step[4][2];
	int steps, freeze, i, nreal;
	int adjfirstbeat;
	char roll[64];
	int rolltime;
	int density;

	if(!fname) {
		l->nstep = 1;
		l->duration = 64;
		l->length = 1;
		fprintf(stderr, "\t(Blank)\n");
		return;
	}

	f = fopen(fname, "r");
	if(!f) err(1, "%s", fname);

	lasttime = 0;
	adjfirstbeat = firstbeat;

	memset(roll, 0, sizeof(roll));
	rolltime = 0;

	while(fgets(buf, sizeof(buf), f)) {
		if(1 == sscanf(buf, "!stop %d", &i)) {
			delay_until(l, now);
			putstepdata(l, 0x10);
			putstepdata(l, i);
			lasttime += i;
			adjfirstbeat += i;
		} else if(6 == sscanf(buf, "%d %3s %1s %1s %1s %1s",
			&now,
			dummy,
			step[0],
			step[1],
			step[2],
			step[3]))
		{
			steps = 0;
			freeze = 0;
			nreal = 0;

			for(i = 0; i < 4; i++) {
				switch(step[i][0]) {
					case '<':
					case 'v':
					case '^':
					case '>':
						nreal++;
						l->nstep++;
						steps |= 1 << i;
						break;
					case 'H':
						nreal++;
						l->nstep++;
						l->nfreeze++;
						freeze |= 1 << i;
						break;
					case '*':
						l->nstep++;
						steps |= 1 << i;
						break;
					case '.':
						break;
					default:
						errx(1, "Unknown step character: '%c'", step[i][0]);
				}
			}

			if(nreal > 1) l->njump++;

			if(nreal) {
				if((now - adjfirstbeat) % beatrate) {
					l->noffbeat++;
				}

				while(rolltime < now) {
					memmove(roll, roll + 1, sizeof(roll) - 1);
					roll[sizeof(roll) - 1] = 0;
					rolltime++;
				}
				roll[sizeof(roll) - 1] = nreal;

				density = 0;
				for(i = 0; i < sizeof(roll); i++) {
					density += roll[i];
				}
				if(density > l->peak) l->peak = density;
			}

			if(steps | freeze) {
				delay_until(l, now);

				if(freeze) {
					putstepdata(l, 0x30);
					putstepdata(l, freeze << 4);
				}

				if(steps) {
					putsteps(l, steps);
				}
			}
		} else if(1 == sscanf(buf, "%d change", &now)) {
			delay_until(l, now);
			putstepdata(l, 0x20);
			putstepdata(l, 0x00);
		}
	}

	delay_until(l, now);
	putstepdata(l, 0x00);

	l->duration = now;

	l->pscore = 1000000 / (l->nstep * (l->nstep + 1) / 2);
	l->qscore = 5 * (1000000 - l->pscore * (l->nstep * (l->nstep + 1) / 2));

	fprintf(stderr, "\tSteps:    %d\n", l->nstep);
	fprintf(stderr, "\tOff-beat: %d\n", l->noffbeat);
	fprintf(stderr, "\tJumps:    %d\n", l->njump);
	fprintf(stderr, "\tPeak:     %d\n", l->peak);
	fprintf(stderr, "\tDuration: %d (%d:%02d)\n", l->duration, l->duration / 50 / 60, (l->duration / 50) % 60);
	fprintf(stderr, "\tP:        %d\n", l->pscore);
	fprintf(stderr, "\tQ:        %d\n", l->qscore);

	if(l->duration > maxduration) maxduration = l->duration;

	fclose(f);
}

int main(int argc, char **argv) {
	FILE *f;
	char buf[256], argbuf[256];
	uint8_t sidbuf[16384];
	int sidsize;
	int i, j, addr;

	char *title = "UNTITLED";
	char *composer = "UNKNOWN";
	char *collection = "HARDSYNC 1ST MIX";
	char *sidfname = "no_filename_given";
	char *levelfname[3] = {0, 0, 0};
	int subtune = 1;
	int bpm;
	int speed = 6;

	if(argc != 2) errx(1, "Usage: %s directory", argv[0]);

	if(chdir(argv[1])) err(1, "%s", argv[1]);

	f = fopen("metadata", "r");
	if(!f) err(1, "metadata");

	while(fgets(buf, sizeof(buf), f)) {
		if(1 == sscanf(buf, "title \"%[^\"]\"", argbuf)) {
			title = strdup(argbuf);
		} else if(1 == sscanf(buf, "composer \"%[^\"]\"", argbuf)) {
			composer = strdup(argbuf);
		} else if(1 == sscanf(buf, "collection \"%[^\"]\"", argbuf)) {
			collection = strdup(argbuf);
		} else if(1 == sscanf(buf, "sid \"%[^\"]\"", argbuf)) {
			sidfname = strdup(argbuf);
		} else if(1 == sscanf(buf, "light \"%[^\"]\"", argbuf)) {
			levelfname[0] = strdup(argbuf);
		} else if(1 == sscanf(buf, "medium \"%[^\"]\"", argbuf)) {
			levelfname[1] = strdup(argbuf);
		} else if(1 == sscanf(buf, "heavy \"%[^\"]\"", argbuf)) {
			levelfname[2] = strdup(argbuf);
		} else if(1 == sscanf(buf, "speed %d", &i)) {
			speed = i;
		} else if(1 == sscanf(buf, "beatrate %d", &i)) {
			beatrate = i;
		} else if(1 == sscanf(buf, "firstbeat %d", &i)) {
			firstbeat = i;
		} else if(1 == sscanf(buf, "subtune %d", &i)) {
			subtune = i;
		}
	}

	fclose(f);

	delaytable[0] = speed * 16;
	delaytable[1] = speed * 16 - 1;
	delaytable[2] = speed * 8;
	delaytable[3] = speed * 8 - 1;
	delaytable[4] = speed * 6;
	delaytable[5] = speed * 6 - 1;
	delaytable[6] = speed * 4;
	delaytable[7] = speed * 4 - 1;
	delaytable[8] = speed * 3;
	delaytable[9] = speed * 3 - 1;
	delaytable[10] = speed * 2;
	delaytable[11] = speed * 2 - 1;
	delaytable[12] = speed * 1;
	delaytable[13] = speed * 1 - 1;
	delaytable[14] = 1;

	for(i = 0; i < 3; i++) {
		fprintf(stderr, "Level %d\n", i + 1);
		readlevel(&level[i], levelfname[i]);
	}

	f = fopen(sidfname, "r");
	if(!f) err(1, "%s", sidfname);
	sidsize = fread(sidbuf, 1, sizeof(sidbuf), f);
	fclose(f);

	outfile = fopen("packed.prg", "wb");
	if(!outfile) err(1, "packed.prg");

	org = 0x8000;

	putbyte('H');
	putbyte('S');
	putbyte(0);
	putbyte(1);

	// Song header, 64 bytes

	putpadstr(title, 20);
	putpadstr(composer, 20);
	putpadstr(collection, 20);

	bpm = round(50 * 60.0 / beatrate);
	fprintf(stderr, "BPM: %d\n", bpm);
	putbcd(bpm, 2);

	putbcd(maxduration / 50 / 60, 1);
	putbcd((maxduration / 50) % 60, 1);

	// Level headers, 3 x 12 bytes

	for(i = 0; i < 3; i++) {
		struct level *l = &level[i];
		int stats[5];

		stats[0] = l->peak * 8 / 10;
		stats[1] = l->nstep * 16 * 8 / l->duration;
		stats[2] = l->njump? (1 + (l->njump - 1) * 8 / 14) : 0;
		stats[3] = l->nfreeze? (1 + (l->nfreeze - 1) * 8 / 14) : 0;
		stats[4] = l->noffbeat * 20 / l->nstep;

		putbyte('-');
		putbyte('-');
		putbyte('-');
		putbyte(0);
		putbyte(0);
		putbyte(0);
		putbyte(0);

		for(j = 0; j < 5; j++) {
			if(stats[j] < 0) stats[j] = 0;
			if(stats[j] > 8) stats[j] = 8;
			putbyte(stats[j]);
		}
	}

	// (100 bytes so far, after load address and magic)

	// Extended song header (not needed for menu), 29 bytes

	fprintf(stderr, "SID size: %d bytes\n", sidsize - 126);

	putword(org + 29 + 3 * 13);			// start of music data
	putword(sidsize - 126);				// length of music data
	putword((sidbuf[125] << 8) | sidbuf[124]);	// load address
	putword((sidbuf[10] << 8) | sidbuf[11]);	// init routine
	putword((sidbuf[12] << 8) | sidbuf[13]);	// play routine
	putbyte(subtune - 1);
	putbyte(beatrate);
	putbyte(firstbeat);
	putbyte(0);
	for(i = 0; i < 15; i++) {
		putbyte(delaytable[i]);
	}

	// Extended level headers (not needed for menu), 3 x 13 bytes

	addr = org + 3 * 13 + sidsize - 126;
	for(i = 0; i < 3; i++) {
		struct level *l = &level[i];

		putword(addr);
		putword(l->nstep);
		putbcd(l->pscore, 4);
		putbcd(l->qscore, 4);
		putbyte(l->initdelay);
		addr += l->length;
	}

	// Data pointed out by previous fields

	fwrite(sidbuf + 126, 1, sidsize - 126, outfile);
	for(i = 0; i < 3; i++) {
		struct level *l = &level[i];

		fwrite(l->data, 1, l->length, outfile);
	}

	fclose(outfile);

	return 0;
}
